from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Tuple

import whylogs_sketching as ds  # type: ignore

# MetricConfig default values

hll_lg_k: int = 12
kll_k: int = 256
fi_lg_max_k: int = 10  # 128 entries
fi_disabled: bool = False
track_unicode_ranges: bool = False
large_kll_k: bool = True
kll_k_large: int = 1024
unicode_ranges: Dict[str, Tuple[int, int]] = {
    "emoticon": (0x1F600, 0x1F64F),
    "control": (0x00, 0x1F),
    "digits": (0x30, 0x39),
    "latin-upper": (0x41, 0x5A),
    "latin-lower": (0x61, 0x7A),
    "basic-latin": (0x00, 0x7F),
    "extended-latin": (0x0080, 0x02AF),
}
lower_case: bool = True  # Convert Unicode characters to lower-case before counting Unicode ranges
normalize: bool = True  # Unicode normalize strings before counting Unicode ranges
max_frequent_item_size: int = 128
identity_column: Optional[str] = None
column_batch_size: Optional[int] = 1024


class FrequentItemsErrorType(int, Enum):
    NO_FALSE_NEGATIVES = 1
    NO_FALSE_POSITIVES = 0

    def to_datasketches_type(self) -> ds.frequent_items_error_type:
        if self.value == FrequentItemsErrorType.NO_FALSE_NEGATIVES.value:
            return ds.frequent_items_error_type.NO_FALSE_NEGATIVES
        else:
            return ds.frequent_items_error_type.NO_FALSE_POSITIVES


@dataclass
class SummaryConfig:
    disabled_metrics: List[str] = field(default_factory=list)
    frequent_items_error_type: FrequentItemsErrorType = FrequentItemsErrorType.NO_FALSE_POSITIVES
    frequent_items_limit: int = -1
    hll_stddev: int = 1



from multiprocessing import Queue
from queue import Empty
from typing import Generic, List, Optional, TypeVar

from whylogs.api.logger.experimental.logger.actor.actor import (
    _DEFAULT_TIMEOUT,
    QueueWrapper,
)

MPWrapperType = TypeVar("MPWrapperType")


class MPQueueWrapper(QueueWrapper, Generic[MPWrapperType]):
    """
    Implementation of QueueWrapper sufficient for use in the threaded actor.
    """

    def __init__(self) -> None:
        self._queue: Queue[MPWrapperType] = Queue(100_000)

    def send(self, message: MPWrapperType, timeout: float = _DEFAULT_TIMEOUT) -> None:
        self._queue.put(message, timeout=timeout)

    def send_many(self, messages: List[MPWrapperType], timeout: float = _DEFAULT_TIMEOUT) -> None:
        for message in messages:
            self._queue.put(message, timeout=timeout)

    def get(self, timeout: float = _DEFAULT_TIMEOUT) -> Optional[MPWrapperType]:
        return self._queue.get(timeout=timeout)

    def get_many(self, timeout: float = _DEFAULT_TIMEOUT, max: Optional[int] = None) -> List[MPWrapperType]:
        if max is None or max < 1:
            return []

        messages: List[MPWrapperType] = []

        while True:
            if len(messages) >= max:
                break

            try:
                messages.append(self._queue.get(timeout=timeout))
            except Empty:
                if len(messages) == 0:
                    raise Empty
                else:
                    break

        return messages

    def size(self) -> int:
        return self._queue.qsize()

    def close(self) -> None:
        self._queue.close()



import configparser
import getpass
import logging
import os
from configparser import ConfigParser
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional

from platformdirs import user_config_dir

from whylogs.api.whylabs.session.notebook_check import is_interractive
from whylogs.api.whylabs.session.prompts import (
    prompt_api_key,
    prompt_default_dataset_id,
    prompt_org_id,
    prompt_session_type,
)
from whylogs.api.whylabs.session.session_types import ApiKeyV1, ApiKeyV2
from whylogs.api.whylabs.session.session_types import InteractiveLogger as il
from whylogs.api.whylabs.session.session_types import (
    SessionType,
    parse_api_key,
    parse_api_key_v2,
)

_DEFAULT_WHYLABS_HOST = "https://api.whylabsapp.com"
_CONFIG_APP_NAME = "whylogs"
_INIT_DOCS = "https://docs.whylabs.ai/docs/whylabs-whylogs-init"


class InitException(Exception):
    pass


class EnvVariableName(Enum):
    WHYLABS_API_KEY = "WHYLABS_API_KEY"
    WHYLABS_SESSION_ID = "WHYLABS_SESSION_ID"
    WHYLABS_ORG_ID = "WHYLABS_DEFAULT_ORG_ID"
    WHYLABS_DEFAULT_DATASET_ID = "WHYLABS_DEFAULT_DATASET_ID"
    WHYLOGS_CONFIG_PATH = "WHYLOGS_CONFIG_PATH"
    WHYLABS_API_ENDPOINT = "WHYLABS_API_ENDPOINT"
    WHYLABS_REFERENCE_PROFILE_NAME = "WHYLABS_REFERENCE_PROFILE_NAME"
    WHYLABS_PRIVATE_API_ENDPOINT = "WHYLABS_PRIVATE_API_ENDPOINT"
    WHYLABS_PRIVATE_S3_ENDPOINT = "WHYLABS_PRIVATE_S3_ENDPOINT"
    HTTP_PROXY = "HTTP_PROXY"
    HTTPS_PROXY = "HTTPS_PROXY"


class ConfigVariableName(Enum):
    API_KEY = "api_key"
    ORG_ID = "org_id"
    SESSION_ID = "session_id"
    WHYLABS_API_ENDPOINT = "whylabs_api_endpoint"
    USER_GUID = "user_guid"
    DEFAULT_DATASET_ID = "default_dataset_id"
    WHYLABS_REFERENCE_PROFILE_NAME = "whylabs_reference_profile_name"
    WHYLABS_PRIVATE_API_ENDPOINT = "whylabs_private_api_endpoint"
    WHYLABS_PRIVATE_S3_ENDPOINT = "whylabs_private_s3_endpoint"
    HTTP_PROXY = "http_proxy"
    HTTPS_PROXY = "https_proxy"


@dataclass
class InitConfig:
    whylabs_api_key: Optional[str] = None
    allow_anonymous: bool = True
    allow_local: bool = False
    default_dataset_id: Optional[str] = None
    config_path: Optional[str] = None
    force_local: Optional[bool] = None


class SessionConfig:
    def __init__(self, init_config: Optional[InitConfig] = None) -> None:
        self._init_config = init_config or InitConfig()
        self.logger = logging.getLogger("config")
        self.auth_path = (
            Path(self._init_config.config_path) if self._init_config.config_path else self.get_config_file_path()
        )
        self._init_parser()

        self.tmp_api_key: Optional[str] = self._init_config.whylabs_api_key
        self.tmp_default_dataset_id: Optional[str] = self._init_config.default_dataset_id

        # Only here for CLI compatibilty. No one should actually set this.
        force_interactive = os.environ.get("WHYLABS_FORCE_INTERACTIVE", "false").lower() == "true"
        if force_interactive:
            self.reset_config()
            self.session_type = self._determine_session_type_prompt(self._init_config)
        else:
            self.session_type = self._determine_session_type(self._init_config)

    def _init_parser(self) -> None:
        try:
            self._config_parser = ConfigFile.create_parser(self.auth_path)
        except (configparser.Error, OSError, IOError, PermissionError) as e:
            self.logger.warning(
                f"Error reading config file from {self.auth_path}: {str(e)}. Operations with WhyLabs will fail."
            )

    def get_session_type(self) -> SessionType:
        return self.session_type

    def _load_value(self, env_name: EnvVariableName, config_name: ConfigVariableName) -> Optional[str]:
        """
        Load a configuration value, first checking the environment, then falling back to the config file
        """
        env_variable = os.getenv(env_name.value)
        if env_variable is not None:
            return env_variable

        return ConfigFile.get_variable_from_config_file(self._config_parser, config_name)

    def _load_or_prompt(
        self, env_name: EnvVariableName, config_name: ConfigVariableName, persist: bool = False, password: bool = True
    ) -> Optional[str]:
        """
        Loads a configuration value like _load_value does, but it will also prompt the user for the value if
        there isn't one in either of those places, assuming we're in a notebook environment. If persist is True,
        the prompt value will also be persisted to the configuration file.
        """
        value = self._load_value(env_name=env_name, config_name=config_name)

        if value is None and is_interractive():
            if password:
                prompt_value = ConfigFile.get_variable_from_getpass(config_name)
            else:
                prompt_value = ConfigFile.get_variable_from_input(config_name)

            if persist:
                ConfigFile.set_variable_to_config_file(self._config_parser, self.auth_path, config_name, prompt_value)
            return prompt_value
        else:
            return value

    def _remove_value(self, key: ConfigVariableName) -> None:
        ConfigFile.remove_variable_from_config_file(self._config_parser, self.auth_path, key)

    def _set_value(self, key: ConfigVariableName, value: str) -> None:
        ConfigFile.set_variable_to_config_file(
            parser=self._config_parser,
            auth_path=self.auth_path,
            key=key,
            value=value,
        )

    def get_https_proxy(self) -> Optional[str]:
        return self._load_value(
            env_name=EnvVariableName.HTTPS_PROXY,
            config_name=ConfigVariableName.HTTPS_PROXY,
        )

    def set_https_proxy(self, proxy: str) -> None:
        self._set_value(ConfigVariableName.HTTPS_PROXY, proxy)

    def get_http_proxy(self) -> Optional[str]:
        return self._load_value(
            env_name=EnvVariableName.HTTP_PROXY,
            config_name=ConfigVariableName.HTTP_PROXY,
        )

    def set_http_proxy(self, proxy: str) -> None:
        self._set_value(ConfigVariableName.HTTP_PROXY, proxy)

    def get_whylabs_private_api_endpoint(self) -> Optional[str]:
        return self._load_value(
            env_name=EnvVariableName.WHYLABS_PRIVATE_API_ENDPOINT,
            config_name=ConfigVariableName.WHYLABS_PRIVATE_API_ENDPOINT,
        )

    def set_whylabs_private_api_endpoint(self, endpoint: str) -> None:
        self._set_value(ConfigVariableName.WHYLABS_PRIVATE_API_ENDPOINT, endpoint)

    def get_whylabs_private_s3_endpoint(self) -> Optional[str]:
        return self._load_value(
            env_name=EnvVariableName.WHYLABS_PRIVATE_S3_ENDPOINT,
            config_name=ConfigVariableName.WHYLABS_PRIVATE_S3_ENDPOINT,
        )

    def set_whylabs_private_s3_endpoint(self, endpoint: str) -> None:
        self._set_value(ConfigVariableName.WHYLABS_PRIVATE_S3_ENDPOINT, endpoint)

    def get_whylabs_refernce_profile_name(self) -> Optional[str]:
        return self._load_value(
            env_name=EnvVariableName.WHYLABS_REFERENCE_PROFILE_NAME,
            config_name=ConfigVariableName.WHYLABS_REFERENCE_PROFILE_NAME,
        )

    def set_whylabs_refernce_profile_name(self, name: str) -> None:
        self._set_value(ConfigVariableName.WHYLABS_REFERENCE_PROFILE_NAME, name)

    def get_whylabs_endpoint(self) -> str:
        return (
            self._load_value(
                env_name=EnvVariableName.WHYLABS_API_ENDPOINT, config_name=ConfigVariableName.WHYLABS_API_ENDPOINT
            )
            or _DEFAULT_WHYLABS_HOST
        )

    def set_whylabs_endpoint(self, endpoint: str) -> None:
        self._set_value(ConfigVariableName.WHYLABS_API_ENDPOINT, endpoint)

    def get_config_file_path(self) -> Path:
        config_dir_path = os.getenv(EnvVariableName.WHYLOGS_CONFIG_PATH.value)
        if config_dir_path is not None:
            Path(config_dir_path).mkdir(parents=True, exist_ok=True)
        else:
            config_dir_path = user_config_dir(_CONFIG_APP_NAME, ensure_exists=True)

        config_file_path = os.path.join(config_dir_path, "config.ini")
        path = Path(config_file_path)
        path.touch(exist_ok=True)
        return path

    def get_default_dataset_id(self) -> Optional[str]:
        return self.tmp_default_dataset_id or self._load_value(
            env_name=EnvVariableName.WHYLABS_DEFAULT_DATASET_ID, config_name=ConfigVariableName.DEFAULT_DATASET_ID
        )

    def require_default_dataset_id(self) -> str:
        return self._require("default dataset id", self.get_default_dataset_id())

    def set_default_dataset_id(self, dataset_id: str) -> None:
        self._set_value(ConfigVariableName.DEFAULT_DATASET_ID, dataset_id)

    def get_org_id(self) -> Optional[str]:
        # Make the v2 api key's org id take precedence over an org id previously stored along with a v1 key
        api_key = self.get_api_key()
        try:
            if api_key is not None:
                return parse_api_key_v2(api_key).org_id
        except Exception:
            pass

        org_id = self._load_value(env_name=EnvVariableName.WHYLABS_ORG_ID, config_name=ConfigVariableName.ORG_ID)

        if org_id is not None:
            return org_id

        return org_id

    def require_org_id(self) -> str:
        return self._require("org id", self.get_org_id())

    def _require(self, name: str, value: Optional[str]) -> str:
        if value is None:
            session_type = self.get_session_type()
            raise ValueError(
                f"Can't determine {name}. Current session type is {session_type.value}. "
                f"See {_INIT_DOCS} for instructions on using why.init()."
            )

        return value

    def set_org_id(self, org_id: str) -> None:
        self._set_value(ConfigVariableName.ORG_ID, org_id)

    def get_api_key(self) -> Optional[str]:
        return self.tmp_api_key or self._load_value(
            env_name=EnvVariableName.WHYLABS_API_KEY, config_name=ConfigVariableName.API_KEY
        )

    def get_env_api_key(self) -> Optional[str]:
        return os.getenv(EnvVariableName.WHYLABS_API_KEY.value)

    def require_api_key(self) -> str:
        return self._require("api key", self.get_api_key())

    def set_api_key(self, api_key: str) -> None:
        self._set_value(ConfigVariableName.API_KEY, api_key)

    def get_user_guid(self) -> Optional[str]:
        return ConfigFile.get_variable_from_config_file(self._config_parser, ConfigVariableName.USER_GUID)

    def set_user_guid(self, user_guid: str) -> None:
        self._set_value(ConfigVariableName.USER_GUID, user_guid)

    def get_session_id(self) -> Optional[str]:
        return self._load_value(env_name=EnvVariableName.WHYLABS_SESSION_ID, config_name=ConfigVariableName.SESSION_ID)

    def set_session_id(self, sessionId: str) -> None:
        self._set_value(ConfigVariableName.SESSION_ID, sessionId)

    def remove_session_id(self) -> None:
        self._remove_value(ConfigVariableName.SESSION_ID)

    def reset_config(self) -> None:
        ConfigFile.remove_all_variables_from_config_file(parser=self._config_parser, auth_path=self.auth_path)
        self._init_parser()

    def notify_session_type(self) -> None:
        config_path = self.get_config_file_path()
        il.message(f"Initializing session with config {config_path}", ignore_suppress=True)
        il.message(ignore_suppress=True)
        if self.session_type == SessionType.WHYLABS:
            self._notify_type_whylabs(self.require_api_key())
        elif self.session_type == SessionType.LOCAL:
            self._notify_type_local()
        elif self.session_type == SessionType.WHYLABS_ANONYMOUS:
            self._notify_type_anon()

    def _notify_type_whylabs(self, api_key: str) -> None:
        default_dataset_id = self.get_default_dataset_id()
        parsed_api_key = parse_api_key(api_key)
        if isinstance(parsed_api_key, ApiKeyV2):
            org_id = parsed_api_key.org_id
        else:
            org_id = self.get_org_id() or "not set"  # Shouldn't be possible to be None at this point

        il.success(f"Using session type: {SessionType.WHYLABS.name}", ignore_suppress=True)
        il.option(f"org id: {org_id}", ignore_suppress=True)
        il.option(f"api key: {parsed_api_key.api_key_id}", ignore_suppress=True)
        if default_dataset_id:
            il.option(f"default dataset: {default_dataset_id}")

        il.message()
        il.message(
            f"In production, you should pass the api key as an environment variable {EnvVariableName.WHYLABS_API_KEY.value}, "
            f"the org id as {EnvVariableName.WHYLABS_ORG_ID.value}, and the default dataset id as "
            f"{EnvVariableName.WHYLABS_DEFAULT_DATASET_ID.value}."
        )

    def _notify_type_anon(self) -> None:
        anonymous_session_id = self.get_session_id()
        il.success(f"Using session type: {SessionType.WHYLABS_ANONYMOUS.name}", ignore_suppress=True)
        id_text = "<will be generated before upload>" if not anonymous_session_id else anonymous_session_id
        il.option(f"session id: {id_text}", ignore_suppress=True)

    def _notify_type_local(self) -> None:
        il.success(
            f"Using session type: {SessionType.LOCAL.name}. "
            "Profiles won't be uploaded or written anywhere automatically.",
            ignore_suppress=True,
        )

    def _determine_session_type_prompt(self, init_config: InitConfig) -> SessionType:
        session_type = prompt_session_type(init_config.allow_anonymous, init_config.allow_local)
        if session_type == SessionType.WHYLABS:
            api_key = prompt_api_key()

            self.set_api_key(api_key.full_key)
            if isinstance(api_key, ApiKeyV1):
                # V1 doesn't include the org id so we have to prompt for it separately
                org_id = prompt_org_id()
                self.set_org_id(org_id)

            # Save the default dataset id if they supplied one
            default_dataset_id = prompt_default_dataset_id()
            if default_dataset_id:
                self.set_default_dataset_id(default_dataset_id)

        return session_type

    def _determine_session_type(self, init_config: InitConfig) -> SessionType:
        if init_config.force_local:
            return SessionType.LOCAL

        # If the user supplied a whylabs api key then use whylabs
        if init_config.whylabs_api_key is not None:
            return SessionType.WHYLABS

        # If there is an api key saved in the config file or in the env then use that for whylabs
        config_api_key = self.get_api_key()
        if config_api_key is not None:
            return SessionType.WHYLABS

        # If there is a session id saved in the config file then use that for an anonymous session
        anonymous_session_id = self.get_session_id()
        if anonymous_session_id is not None and init_config.allow_anonymous:
            return SessionType.WHYLABS_ANONYMOUS

        # If we're in an interactive environment then prompt the user to pick an authentication method
        if is_interractive():
            return self._determine_session_type_prompt(init_config)

        # If we're not interactive then pick between anonymous and local based on the init config
        if init_config.allow_anonymous:
            return SessionType.WHYLABS_ANONYMOUS

        if init_config.allow_local:
            return SessionType.LOCAL

        raise InitException(
            "Don't know how to initialize authentication because allow_anonymous=False, allow_local=False, "
            "and there is no WhyLabs api key in the environment, config file, or why.init() call, and this isn't an "
            f"interactive environment. See {_INIT_DOCS} for instructions on using why.init()."
        )


_CONFIG_WHYLABS_SECTION = "whylabs"


class ConfigFile:
    @staticmethod
    def create_parser(config_path: Path) -> ConfigParser:
        config_parser = ConfigParser()
        config_parser.read(config_path)
        if not config_parser.has_section(_CONFIG_WHYLABS_SECTION):
            config_parser[_CONFIG_WHYLABS_SECTION] = {}
        return config_parser

    @staticmethod
    def get_variable_from_input(variable_name: ConfigVariableName) -> str:
        variable = input(f"What is your {variable_name.value}? ")
        if not variable:
            raise ValueError(f"Interrupted before input was entered for {variable_name.value}")
        return variable

    @staticmethod
    def get_variable_from_getpass(variable_name: ConfigVariableName) -> str:
        variable = getpass.getpass(f"What is your {variable_name.value}? ")
        if not variable:
            raise ValueError(f"Interrupted before input was entered for {variable_name.value}")
        return variable

    @staticmethod
    def get_variable_from_config_file(parser: ConfigParser, key: ConfigVariableName) -> Optional[str]:
        value = parser.get(_CONFIG_WHYLABS_SECTION, key.value, fallback=None)
        return value

    @staticmethod
    def set_variable_to_config_file(parser: ConfigParser, auth_path: Path, key: ConfigVariableName, value: str) -> None:
        parser.set(_CONFIG_WHYLABS_SECTION, key.value, value)
        with open(auth_path, "w") as configfile:
            parser.write(configfile)

    @staticmethod
    def remove_variable_from_config_file(parser: ConfigParser, auth_path: Path, key: ConfigVariableName) -> None:
        parser.remove_option(_CONFIG_WHYLABS_SECTION, key.value)
        with open(auth_path, "w") as configfile:
            parser.write(configfile)

    @staticmethod
    def remove_all_variables_from_config_file(parser: ConfigParser, auth_path: Path) -> None:
        parser.remove_section(_CONFIG_WHYLABS_SECTION)
        with open(auth_path, "w") as configfile:
            parser.write(configfile)



import dataclasses
import json
from datetime import datetime, timezone
from functools import partial, reduce
from logging import getLogger
from typing import Dict, Iterable, List, Optional, Tuple

import whylogs as why
from whylogs.api.logger.result_set import ResultSet, SegmentedResultSet
from whylogs.api.pyspark.experimental.profiler import COL_NAME_FIELD, COL_PROFILE_FIELD
from whylogs.api.usage_stats import emit_usage
from whylogs.core import DatasetSchema
from whylogs.core.segment import Segment
from whylogs.core.segmentation_partition import SegmentationPartition
from whylogs.core.stubs import pd
from whylogs.core.view.column_profile_view import ColumnProfileView
from whylogs.core.view.dataset_profile_view import DatasetProfileView

logger = getLogger(__name__)
emit_usage("pyspark")

try:  # type: ignore
    from pyspark.ml.functions import vector_to_array
    from pyspark.ml.linalg import VectorUDT
    from pyspark.sql import DataFrame as SparkDataFrame  # types: ignore
except ImportError:  # noqa
    logger.error("No pyspark available")
    SparkDataFrame = None  # type: ignore

SEGMENT_KEY_FIELD = "segment_key"


def _segment_to_string(segment: Segment) -> str:
    return json.dumps(dataclasses.asdict(segment))


def _string_to_segment(segment_key: str) -> Optional[Segment]:
    if len(segment_key) == 0:
        return None
    segment_params = json.loads(segment_key)
    if "key" not in segment_params or "parent_id" not in segment_params:
        raise ValueError("Segment key is missing required key values or parent_id: {segment_params}")
    tuple_key = tuple(segment_params["key"])
    parent_id = segment_params["parent_id"]
    return Segment(key=tuple_key, parent_id=parent_id)


def whylogs_pandas_segmented_profiler(
    pdf_iterator: Iterable[pd.DataFrame], schema: Optional[DatasetSchema] = None
) -> Iterable[pd.DataFrame]:
    if schema is None or not schema.segments:
        raise ValueError(
            "Cannot profile segments without segmentation defined in the specified DatasetSchema: no segments found."
        )
    for input_df in pdf_iterator:
        # TODO: optimize this so we split the dataframe by segment first rather than split in the pandas dataframe
        segmented_results = why.log(input_df, schema=schema)
        res_df = pd.DataFrame(columns=[SEGMENT_KEY_FIELD, COL_NAME_FIELD, COL_PROFILE_FIELD])
        for segmented_key in segmented_results.segments():
            spark_segment_key = _segment_to_string(segmented_key)
            for col_name, col_profile in segmented_results.view(segmented_key).get_columns().items():
                d = {
                    SEGMENT_KEY_FIELD: [spark_segment_key],
                    COL_NAME_FIELD: [col_name],
                    COL_PROFILE_FIELD: [col_profile.to_protobuf().SerializeToString()],
                }
                df_temp = pd.DataFrame(data=d)
                res_df = pd.concat([res_df, df_temp])
        yield res_df


def column_profile_bytes_aggregator(group_by_cols: Tuple[str], profiles_df: pd.DataFrame) -> pd.DataFrame:
    if len(profiles_df) == 0:
        return pd.DataFrame([group_by_cols + (bytearray(),)])

    merged_profile: ColumnProfileView = reduce(
        lambda acc, x: acc.merge(x), profiles_df[COL_PROFILE_FIELD].apply(ColumnProfileView.from_bytes)
    )
    return pd.DataFrame([group_by_cols + (merged_profile.serialize(),)])


def collect_segmented_column_profile_views(
    input_df: SparkDataFrame, schema: Optional[DatasetSchema] = None
) -> Dict[Segment, Dict[str, ColumnProfileView]]:
    if SparkDataFrame is None:
        logger.warning("Unable to load pyspark; install pyspark to get whylogs profiling support in spark environment.")
    vector_columns = [
        col_name for col_name in input_df.schema.names if isinstance(input_df.schema[col_name].dataType, VectorUDT)
    ]
    input_df_arrays = input_df
    for col_name in vector_columns:
        input_df_arrays = input_df_arrays.withColumn(col_name, vector_to_array(input_df_arrays[col_name]))

    cp = f"{SEGMENT_KEY_FIELD} string, {COL_NAME_FIELD} string, {COL_PROFILE_FIELD} binary"
    whylogs_pandas_map_profiler_with_schema = partial(whylogs_pandas_segmented_profiler, schema=schema)
    segmented_profile_bytes_df = input_df_arrays.mapInPandas(whylogs_pandas_map_profiler_with_schema, schema=cp)  # type: ignore
    # aggregate by segment key first, then by column
    segment_column_profiles = segmented_profile_bytes_df.groupby(
        SEGMENT_KEY_FIELD, COL_NAME_FIELD
    ).applyInPandas(  # linebreak
        column_profile_bytes_aggregator, schema=cp
    )

    collected_segment_column_profile_views: Dict[Tuple[Segment, str], ColumnProfileView] = {
        (_string_to_segment(row.segment_key), row.col_name): ColumnProfileView.from_bytes(row.col_profile)
        for row in segment_column_profiles.collect()
    }

    segmented_column_profile_views: Dict[Segment, Dict[str, ColumnProfileView]] = dict()

    for key_tuple in collected_segment_column_profile_views:
        segment_key = key_tuple[0]
        column_name = key_tuple[1]
        if segment_key not in segmented_column_profile_views:
            segmented_column_profile_views[segment_key] = dict()
        segmented_column_profile_views[segment_key][column_name] = collected_segment_column_profile_views[key_tuple]

    return segmented_column_profile_views


def _lookup_segment_partition_by_id(schema: DatasetSchema, partition_id: str) -> Optional[SegmentationPartition]:
    for partition_name in schema.segments:
        partition = schema.segments[partition_name]
        if partition_id == partition.id:
            return partition
    logger.warning(f"Skipping segment: could not find partition with id {partition_id} in: {schema.segments}")
    return None


def collect_segmented_results(
    input_df: SparkDataFrame,
    schema: DatasetSchema,
    dataset_timestamp: Optional[int] = None,
    creation_timestamp: Optional[int] = None,
) -> ResultSet:
    now = datetime.now(timezone.utc)

    _dataset_timestamp = dataset_timestamp or now
    _creation_timestamp = creation_timestamp or now
    if not schema.segments:
        raise ValueError("Cannot collect segmented results without segments defined in the passed in DatasetSchema")

    logger.info(f"Processing segmented profiling in spark with {schema.segments}")
    segment_column_views_dict = collect_segmented_column_profile_views(input_df=input_df, schema=schema)
    segments: Dict[str, Dict[Segment, DatasetProfileView]] = dict()
    partitions: List[SegmentationPartition] = list(schema.segments.values())

    for segment in segment_column_views_dict:
        partition_id = segment.parent_id
        # check that the segment's parent id has a matching partition
        partition = _lookup_segment_partition_by_id(schema, partition_id)
        if partition is None:
            logger.error(
                f"Skipping segment: could not collect profiles for segment {segment} because the schema has no matching partition ids: {schema.segments.keys()}"
            )
            continue
        if partition_id not in segments:
            segments[partition_id] = dict()
        if segment not in segments[partition_id]:
            column_views_dict = segment_column_views_dict[segment]
            segments[partition_id][segment] = DatasetProfileView(
                columns=column_views_dict, dataset_timestamp=_dataset_timestamp, creation_timestamp=_creation_timestamp
            )
        else:
            raise ValueError(f"Collision when collecting profiles for segment {segment}!")

    return SegmentedResultSet(segments=segments, partitions=partitions)



import logging
from copy import deepcopy
from dataclasses import dataclass, field
from itertools import chain
from typing import Any, Dict, List, Optional, Set, Union

import whylogs as why
from whylogs.api.logger.result_set import ResultSet
from whylogs.core.datatypes import (
    DataType,
    Fractional,
    Integral,
    StandardTypeMapper,
    String,
    TypeMapper,
)
from whylogs.core.metrics import StandardMetric
from whylogs.core.metrics.metrics import (
    Metric,
    MetricConfig,
    OperationResult,
    register_metric,
)
from whylogs.core.metrics.multimetric import MultiMetric, SubmetricSchema
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.resolvers import Resolver
from whylogs.core.schema import ColumnSchema, DatasetSchema

logger = logging.getLogger(__name__)

try:
    import numpy as np  # type: ignore
    from PIL import Image
    from PIL.Image import Image as ImageType  # type: ignore
    from PIL.ImageStat import Stat  # type: ignore
    from PIL.TiffImagePlugin import IFDRational  # type: ignore
    from PIL.TiffTags import TAGS  # type: ignore
except ImportError as e:
    ImageType = None  # type: ignore
    logger.warning(str(e))
    logger.warning("Unable to load PIL; install Pillow for image support")

DEFAULT_IMAGE_FEATURES: List[str] = []


_DEFAULT_TAGS_ATTRIBUTES = [
    "ImagePixelWidth",
    "ImagePixelHeight",
    "Colorspace",
]
_IMAGE_HSV_CHANNELS = ["Hue", "Saturation", "Brightness"]
_STATS_PROPERTIES = ["mean", "stddev"]
_DEFAULT_STAT_ATTRIBUTES = [c + "." + s for c in _IMAGE_HSV_CHANNELS for s in _STATS_PROPERTIES]
_METADATA_DEFAULT_ATTRIBUTES = _DEFAULT_STAT_ATTRIBUTES + _DEFAULT_TAGS_ATTRIBUTES


def get_pil_image_statistics(
    img: ImageType, channels: List[str] = _IMAGE_HSV_CHANNELS, image_stats: List[str] = _STATS_PROPERTIES
) -> Dict:
    """
    Compute statistics data for a PIL Image

    Args:
        img (ImageType): PIL Image

    Returns:
        Dict: of metadata
    """

    stats = Stat(img.convert("HSV"))
    metadata = {}
    if hasattr(img, "entropy"):
        metadata["entropy"] = img.entropy()
    for index in range(len(channels)):
        for statistic_name in image_stats:
            if hasattr(stats, statistic_name):
                metadata[channels[index] + "." + statistic_name] = getattr(stats, statistic_name)[index]

    return metadata


def get_pil_exif_metadata(img: ImageType) -> Dict:
    """
    Grab EXIF metadata from image

    Args:
        img (ImageType): PIL Image

    Returns:
        Dict: of metadata
    """
    metadata = {}
    for k, v in dict(img.getexif()).items():
        try:
            if isinstance(v, IFDRational):
                metadata[TAGS[k]] = "{}".format(v)
            else:
                metadata[TAGS[k]] = v
        except KeyError:
            logger.warning(f"Couldn't read exif tag: {k} skipping.")

    return metadata


def image_based_metadata(img: ImageType) -> Dict[str, int]:
    return {
        "ImagePixelWidth": img.width,
        "ImagePixelHeight": img.height,
        "Colorspace": img.mode,
    }


def get_pil_image_metadata(img: ImageType) -> Dict:
    """
    Grab statistics data from a PIL ImageStats.Stat

    Args:
        img (ImageType): PIL Image

    Returns:
        Dict: of metadata
    """
    metadata = {}
    metadata.update(image_based_metadata(img))
    metadata.update(get_pil_image_statistics(img))
    return metadata


class ImageSubmetricSchema(SubmetricSchema):
    def resolve(self, name: str, why_type: DataType, fi_disabled: bool = False) -> Dict[str, Metric]:
        metrics: Dict[str, Metric] = {
            "counts": StandardMetric.counts.zero(MetricConfig()),
            "types": StandardMetric.types.zero(MetricConfig()),
            "cardinality": StandardMetric.cardinality.zero(MetricConfig()),
        }

        if isinstance(why_type, Integral):
            metrics["distribution"] = StandardMetric.distribution.zero(MetricConfig())
            metrics["ints"] = StandardMetric.ints.zero(MetricConfig())
            if not fi_disabled:
                metrics["frequent_items"] = StandardMetric.frequent_items.zero(MetricConfig())
        elif isinstance(why_type, Fractional):
            metrics["distribution"] = StandardMetric.distribution.zero(MetricConfig())
        elif isinstance(why_type, String) and not fi_disabled:
            metrics["frequent_items"] = StandardMetric.frequent_items.zero(MetricConfig())

        return metrics


@dataclass(frozen=True)
class ImageMetricConfig(MetricConfig):
    allowed_exif_tags: Set[str] = field(default_factory=set)
    forbidden_exif_tags: Set[str] = field(default_factory=set)


class ImageMetric(MultiMetric):
    def __init__(
        self,
        submetrics: Dict[str, Dict[str, Metric]],
        allowed_exif_tags: Optional[Set[str]] = None,
        forbidden_exif_tags: Optional[Set[str]] = None,
        submetric_schema: Optional[SubmetricSchema] = None,
        type_mapper: Optional[TypeMapper] = None,
        fi_disabled: bool = False,
    ):
        if ImageType is None:
            logger.error("Install Pillow for image support")
        super(ImageMetric, self).__init__(submetrics)
        self._allowed_exif_tags = allowed_exif_tags or set()
        self._forbidden_exif_tags = forbidden_exif_tags or set()
        self._submetric_schema = submetric_schema or ImageSubmetricSchema()
        self._type_mapper = type_mapper or StandardTypeMapper()
        self._fi_disabled = fi_disabled

    @property
    def namespace(self) -> str:
        return "image"

    def merge(self, other: "ImageMetric") -> "ImageMetric":
        merged = super(ImageMetric, self).merge(other)
        merged._allowed_exif_tags = self._allowed_exif_tags.copy()
        merged._forbidden_exif_tags = self._forbidden_exif_tags.copy()
        merged._submetric_schema = deepcopy(self._submetric_schema)
        merged._type_mapper = deepcopy(self._type_mapper)
        merged._fi_disabled = self._fi_disabled
        return merged

    def _wants_to_track(self, exif_tag: str) -> bool:
        if exif_tag in self._forbidden_exif_tags:
            return False
        if not self._allowed_exif_tags:  # empty set -> allow anything not explicitly forbidden
            return True
        return exif_tag in self._allowed_exif_tags

    def _add_submetric(self, name: str, value: Any) -> None:
        self.submetrics[name] = self._submetric_schema.resolve(name, self._type_mapper(type(value)), self._fi_disabled)

    def _discover_exif_submetrics(self, name: str, value: Any) -> None:
        if name not in self.submetrics and self._wants_to_track(name):
            self._add_submetric(name, value)

    def _discover_image_submetrics(self, name: str, value: Any) -> None:
        if name not in self.submetrics:
            self._add_submetric(name, value)

    def _update_relevant_submetrics(self, name: str, data: PreprocessedColumn) -> None:
        if name not in self.submetrics:
            return

        for metric in self.submetrics[name].values():
            metric.columnar_update(data)

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        count = 0
        for image in list(chain.from_iterable(view.raw_iterator())):
            if isinstance(image, np.ndarray):
                image = Image.fromarray(image.astype(np.uint8))

            if isinstance(image, ImageType):
                metadata = get_pil_exif_metadata(image)
                for name, value in metadata.items():
                    self._discover_exif_submetrics(name, value)  # EXIF tag discovery
                    data = PreprocessedColumn.apply([metadata[name]])  # TODO: _process_scalar_value()?
                    self._update_relevant_submetrics(name, data)

                image_data = get_pil_image_metadata(image)
                for name, value in image_data.items():
                    self._discover_image_submetrics(name, value)
                    data = PreprocessedColumn.apply([image_data[name]])
                    self._update_relevant_submetrics(name, data)

                count += 1
        return OperationResult.ok(count)

    @classmethod
    def zero(cls, config: Optional[MetricConfig] = None) -> "ImageMetric":
        config = config or ImageMetricConfig()
        if not isinstance(config, ImageMetricConfig):
            logger.error("ImageMetric.zero() needs an ImageMetricConfig")
            config = ImageMetricConfig()

        return ImageMetric(
            {},  # submetrics will be discovered as images are logged
            config.allowed_exif_tags,
            config.forbidden_exif_tags,
            None,  # use standard ImageSubmetricSchema
            None,  # use standard TypeMapper
            config.fi_disabled,
        )


def init_image_schema(column_prefix: str = "image") -> DatasetSchema:
    """
    Initialize a DatasetSchema for logging images. This can be passed into a logger or why.log.

    Args:
        column_prefix (str): The prefix that appears in the dataset profiles along with all of the
        image features. If the prefix is "image", then you'll log image with why.log({image: image_data}).
    """

    class ImageResolver(Resolver):
        def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
            return {ImageMetric.get_namespace(): ImageMetric.zero(column_schema.cfg)}

    return DatasetSchema(
        types={column_prefix: Image.Image}, default_configs=ImageMetricConfig(), resolvers=ImageResolver()
    )


def log_image(
    images: Union[ImageType, List[ImageType], Dict[str, ImageType]],
    default_column_prefix: str = "image",
    schema: Optional[DatasetSchema] = None,
    trace_id: Optional[str] = None,
) -> ResultSet:
    if isinstance(images, ImageType):
        images = {default_column_prefix: images}
    if isinstance(images, list):
        count = 0
        row = dict()
        for img in images:
            row[f"{default_column_prefix}_{count}"] = img
            count += 1
        images = row

    if not isinstance(images, dict):
        raise ValueError("log_image must be passed an image, list of images, or dictionary of images")
    for img in images.values():
        if not isinstance(img, ImageType):
            raise ValueError("log_image must be passed an image, list of images, or dictionary of images")

    class ImageResolver(Resolver):
        def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
            return {ImageMetric.get_namespace(): ImageMetric.zero(column_schema.cfg)}

    schema = schema or DatasetSchema(
        types={key: ImageType for key in images.keys()}, default_configs=ImageMetricConfig(), resolvers=ImageResolver()
    )
    if not isinstance(schema.default_configs, ImageMetricConfig):
        raise ValueError("log_image requires DatasetSchema with an ImageMetricConfig as default_configs")

    return why.log(row=images, schema=schema, trace_id=trace_id)


# Register it so Multimetric and ProfileView can deserialize
register_metric(ImageMetric)



import abc
import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple

from urllib3 import ProxyManager
from whylabs_client import ApiClient, Configuration  # type: ignore

from whylogs.core.utils.utils import get_auth_headers


class KeyRefresher(abc.ABC):
    @property
    @abc.abstractmethod
    def key_id(self) -> str:
        pass

    @abc.abstractmethod
    def __call__(self, config: Configuration) -> None:
        pass

    def _validate_api_key(self, api_key: Optional[str]) -> str:
        if api_key is None:
            raise ValueError("Missing API key. Set it via WHYLABS_API_KEY environment variable or as an api_key option")
        if len(api_key) < 12:
            raise ValueError("API key too short")
        if len(api_key) > 80:
            raise ValueError("API key too long")
        if api_key[10] != ".":
            raise ValueError("Invalid format. Expecting a dot at an index 10")
        return api_key[:10]


class StaticKeyRefresher(KeyRefresher):
    def __init__(self, api_key: str) -> None:
        self._key_id = self._validate_api_key(api_key)
        self._api_key = api_key

    @property
    def key_id(self) -> str:
        return self._key_id

    def __call__(self, config: Configuration) -> None:
        config.api_key = {"ApiKeyAuth": self._api_key}

    def __hash__(self) -> int:
        return hash(self._api_key)


class EnvironmentKeyRefresher(KeyRefresher):
    """
    This key refresher uses environment variable key. The key is automatically picked up if the
    user changes the environment variable.
    """

    @property
    def key_id(self) -> str:
        return self._key_id

    def __call__(self, config: Configuration) -> None:
        from whylogs.api.whylabs.session.session_manager import _default_init

        session = _default_init()
        session_config = session.config
        api_key = session_config.get_env_api_key()
        self._key_id = self._validate_api_key(api_key)
        assert api_key is not None
        config.api_key = {"ApiKeyAuth": api_key}


@dataclass(frozen=True)
class ClientCacheConfig:
    api_key: Optional[str] = None
    ssl_ca_cert: Optional[str] = None
    whylabs_api_endpoint: Optional[str] = None
    endpoint_hostname: Optional[str] = None  # TODO What is this thing?


class WhylabsClientCache:
    __instance: Optional["WhylabsClientCache"] = None

    # Must be initialized from the session code as a side effect of why.init()
    @staticmethod
    def __init_instance() -> None:
        if WhylabsClientCache.__instance is None:
            WhylabsClientCache.__instance = WhylabsClientCache()

    @staticmethod
    def reset() -> None:
        WhylabsClientCache.__instance = None

    @staticmethod
    def instance() -> "WhylabsClientCache":
        if WhylabsClientCache.__instance is None:
            # Internally, we'll call _default_init() to initialize the instance wherever we need
            # a guarantee that one exists.
            raise ValueError("WhylabsClientCache is not initialized. Call why.init() to initialize it.")
        return WhylabsClientCache.__instance

    def __init__(self) -> None:
        self._api_client_cache: Dict[ClientCacheConfig, Tuple[ApiClient, KeyRefresher]] = dict()
        self._logger = logging.getLogger(__name__)

    def get_client(self, config: ClientCacheConfig) -> Tuple[ApiClient, KeyRefresher]:
        entry = self._api_client_cache.get(config)

        if entry is None:
            entry = self._create_client(config)
            self._api_client_cache[config] = entry

        return entry

    def _create_client(self, cache_config: ClientCacheConfig) -> Tuple[ApiClient, KeyRefresher]:
        """
        Refresh the API client by comparing various configs. We try to
        re-use the client as much as we can since using a new client
        every time can be expensive.
        """
        from whylogs.api.whylabs.session.session_manager import _default_init

        session = _default_init()
        session_config = session.config
        self._proxy = session_config.get_https_proxy() or session_config.get_http_proxy()

        config = Configuration()
        config.api_key = {"ApiKeyAuth": ""}
        refresher: KeyRefresher = (
            StaticKeyRefresher(cache_config.api_key)
            if cache_config.api_key
            else EnvironmentKeyRefresher()  # TODO when would this be used? I'm requiring api keys right now in the whylabs writer
        )
        config.refresh_api_key_hook = refresher
        config.discard_unknown_keys = True
        config.client_side_validation = False  # Disable client side validation and trust the server

        if self._proxy:
            config.proxy = self._proxy
            default_header = get_auth_headers(self._proxy)
            if default_header:
                config.proxy_headers = default_header

        if cache_config.ssl_ca_cert:
            config.ssl_ca_cert = cache_config.ssl_ca_cert

        config.host = cache_config.whylabs_api_endpoint or session_config.get_whylabs_endpoint()

        client = ApiClient(config)
        from whylogs import __version__ as _version

        client.user_agent = f"whylogs/python/{_version}"

        if cache_config.endpoint_hostname:
            self._logger.info(
                f"Override endpoint hostname for TLS verification is set to: {cache_config.endpoint_hostname}"
            )
            self._update_hostname_config(client, cache_config.endpoint_hostname)

        return client, refresher

    def _update_hostname_config(self, client: ApiClient, endpoint_hostname_override: str) -> None:
        """
        This method overrides the pool manager's new connection method to add the hostname
        """
        if isinstance(client.rest_client.pool_manager, ProxyManager):
            raise ValueError("Endpoint hostname override is not supported when using with proxy")

        self._logger.debug(f"Override endpoint hostname to: {endpoint_hostname_override}")
        old_conn_factory = client.rest_client.pool_manager.connection_from_host

        def new_conn_factory(host: str, port: int, scheme: str, pool_kwargs: Optional[Dict[str, str]] = None) -> Any:
            if pool_kwargs is None:
                pool_kwargs = {}
            pool_kwargs["assert_hostname"] = endpoint_hostname_override
            pool_kwargs["server_hostname"] = endpoint_hostname_override
            return old_conn_factory(host, port, scheme, pool_kwargs)

        client.rest_client.pool_manager.connection_from_host = new_conn_factory



# Various common types to avoid circular dependencies
import os
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Optional, Set, Union

from whylogs.api.whylabs.session.notebook_check import is_interractive


# This is used to indicate that a result is either a success or a failure
class NotSupported:
    """
    Indicates that one of the session apis isn't supported for the current session type.
    """


class InteractiveLogger:
    """
    Logger for interactive environments.
    """

    __warnings: Set[int] = set()
    _is_notebook: bool = False

    @staticmethod
    def init_notebook_logging() -> None:
        if is_interractive():
            InteractiveLogger._is_notebook = True

    @staticmethod
    def __should_log(ignore_suppress: bool = False) -> bool:
        """
        Returns true if we should log, false otherwise.
        """
        if ignore_suppress:
            return InteractiveLogger._is_notebook
        else:
            return not os.environ.get("WHYLOGS_SUPPRESS_LOG_OUTPUT") and InteractiveLogger._is_notebook

    @staticmethod
    def message(message: str = "", log_fn: Optional[Callable] = None, ignore_suppress: bool = False) -> None:
        """
        Log a message only if we're in a notebook environment.

        Args:
            message: The message to log
            log_fn: A function to log to instead of printing if we're not in a notebook.
            ignore_suppress: If true, will log even if WHYLOGS_SUPPRESS_LOG_OUTPUT is set. It still needs
                to be in a notebook though or it won't show.
        """
        if InteractiveLogger.__should_log(ignore_suppress=ignore_suppress):
            print(message)
        elif log_fn is not None:
            log_fn(message)

    @staticmethod
    def option(message: str, ignore_suppress: bool = False) -> None:
        """
        Log an option line, which is anything that has multiple related lines in a row like
        multiple choices or a list things.
        """
        InteractiveLogger.message(f" ⤷ {message}", ignore_suppress=ignore_suppress)

    @staticmethod
    def inspect(message: str) -> None:
        """
        Log a message that the user is intended to interact with or inspect, like a url
        """
        InteractiveLogger.message(f"🔍 {message}")

    @staticmethod
    def question(message: str, ignore_suppress: bool = False) -> None:
        """
        Log a question.
        """
        InteractiveLogger.message(f"❓ {message}", ignore_suppress=ignore_suppress)

    @staticmethod
    def success(message: str, ignore_suppress: bool = False) -> None:
        """
        Log a success line, which has a green checkmark.
        """
        InteractiveLogger.message(f"✅ {message}", ignore_suppress=ignore_suppress)

    @staticmethod
    def failure(message: str, ignore_suppress: bool = False) -> None:
        """
        Log a failure, which has a red x.
        """
        InteractiveLogger.message(f"❌ {message}", ignore_suppress=ignore_suppress)

    @staticmethod
    def warning(message: str, log_fn: Optional[Callable] = None, ignore_suppress: bool = False) -> None:
        """
        Log a warning, which has a warning sign.
        """
        InteractiveLogger.message(f"⚠️ {message}", log_fn=log_fn, ignore_suppress=ignore_suppress)

    @staticmethod
    def warning_once(message: str, log_fn: Optional[Callable] = None, ignore_suppress: bool = False) -> None:
        """
        Like warning, but only logs once.
        """
        if not InteractiveLogger.__should_log(ignore_suppress=ignore_suppress):
            return

        if hash(message) not in InteractiveLogger.__warnings:
            InteractiveLogger.message(f"⚠️ {message}", log_fn=log_fn)
            InteractiveLogger.__warnings.add(hash(message))


InteractiveLogger.init_notebook_logging()


class SessionType(Enum):
    WHYLABS_ANONYMOUS = "whylabs_anonymous"
    WHYLABS = "whylabs"
    LOCAL = "local"


@dataclass
class ApiKeyV1:
    api_key_id: str
    full_key: str


@dataclass
class ApiKeyV2:
    api_key_id: str
    org_id: str
    full_key: str


ApiKey = Union[ApiKeyV1, ApiKeyV2]


def parse_api_key(api_key: str) -> ApiKey:
    try:
        return parse_api_key_v2(api_key)
    except Exception:
        return parse_api_key_v1(api_key)


def validate_org_id(org_id: str) -> None:
    org_id.startswith("org-")


def parse_api_key_v1(api_key: str) -> ApiKeyV1:
    """
    Parse the key id from an api key.
    V1 api keys have the format key_id.key
    """

    # split on . and get the first part
    api_key_id = api_key.split(".")[0]

    assert ":" not in api_key
    return ApiKeyV1(api_key_id, api_key)


def parse_api_key_v2(api_key: str) -> ApiKeyV2:
    """
    Parse the key id and the org id from an api key
    V2 api keys have the format key_id.key:org_id
    """

    # split on . and get the first part
    api_key_id = api_key.split(".")[0]

    # split on : and get the second part
    org_id = api_key.split(":")[1]
    return ApiKeyV2(api_key_id, org_id, api_key)



import logging
from dataclasses import dataclass
from typing import Any

logger = logging.getLogger(__name__)

try:
    import pandas as _pd
except ImportError:  # noqa
    _pd = None  # type: ignore

try:
    import numpy as _np
except ImportError:  # noqa
    _np = None  # type: ignore
    if _pd is not None:
        logger.error("Pandas is installed but numpy is not. Your environment is probably broken.")

try:
    import scipy as _sp
except ImportError:  # noqa
    _sp = None  # type: ignore

try:
    import sklearn.cluster as _sklc
    import sklearn.decomposition as _skld
    import sklearn.metrics.pairwise as _sklp
except ImportError:  # noqa
    _sklp = None  # type: ignore
    _sklc = None  # type: ignore
    _skld = None  # type: ignore


class _StubClass:
    pass


@dataclass(frozen=True)
class NumpyStub:
    dtype: type = _StubClass
    number: type = _StubClass
    bool_: type = _StubClass
    floating: type = _StubClass
    ndarray: type = _StubClass
    timedelta64: type = _StubClass
    datetime64: type = _StubClass
    unicode_: type = _StubClass
    issubdtype: type = _StubClass
    integer: type = _StubClass


@dataclass(frozen=True)
class PandasStub(object):
    Series: type = _StubClass
    DataFrame: type = _StubClass


@dataclass(frozen=True)
class ScipyStub:
    sparse: type = _StubClass


@dataclass(frozen=True)
class ScikitLearnStub:
    cosine_distances: type = _StubClass
    euclidean_distances: type = _StubClass
    KMeans: type = _StubClass
    PCA: type = _StubClass


def is_not_stub(stubbed_class: Any) -> bool:
    if (
        stubbed_class
        and stubbed_class is not _StubClass
        and not isinstance(stubbed_class, (PandasStub, NumpyStub, ScipyStub, ScikitLearnStub))
    ):
        return True
    return False


if _np is None:
    _np = NumpyStub()

if _pd is None:
    _pd = PandasStub()

if _sp is None:
    _sp = ScipyStub()

if _sklp is None:
    _sklp = ScikitLearnStub()

if _sklc is None:
    _sklc = ScikitLearnStub()

if _skld is None:
    _skld = ScikitLearnStub()


np = _np
pd = _pd
sp = _sp
sklp = _sklp
sklc = _sklc
skld = _skld
